//-----------------------------------------------------------------------------
/*
Programmed by Olivier Sangala
http://olivier.sangala.free.fr
olivier_sangala@hotmail.com
*/
//-----------------------------------------------------------------------------

// C Source File
// Created 15/12/01; 13:41:51

#define USE_TI89              // Compile for TI-89
#define USE_TI92PLUS          // Compile for TI-92 Plus
#define USE_V200              // Compile for V200

// #define OPTIMIZE_ROM_CALLS // Use ROM Call Optimization

#define MIN_AMS 100           // Compile for AMS 1.00 or higher

#define SAVE_SCREEN           // Save/Restore LCD Contents

#include <tigcclib.h>         // Include All Header Files

#include "extgraph.h"
#include "graphic.h"
#include "menulib.h"
#include "sokoban.h"

//--------------------------------------------------------------------------

extern TMenu * Menu;
enum {ME_MAIN};

//--------------------------------------------------------------------------

char TS_FLOOR=' ';
char TS_WALL='#';
char TS_GOAL='.';
char TS_BALL='$';
char TS_BALL_INGOAL='*';
char TS_SOKOBAN='@';
char TS_SOKOBAN_INGOAL='+';

//--------------------------------------------------------------------------

char * Plane0;
char * Plane1;

TGame*Game;

int ShowTileOnScreen;

//--------------------------------------------------------------------------

// Custom String Input Function
void InputStr (char *buffer, short maxlen)
{
	SCR_STATE ss;
	short key, i = 0;
	buffer[0] = 0;
	SaveScrState (&ss);
	do
	{
		MoveTo (ss.CurX, ss.CurY);
		printf ("%s_  ", buffer);
		// Note that two spaces are required if F_4x6 font is used
		key = ngetchx ();
		if (key >= ' ' && key <= '~' && i < maxlen) buffer[i++] = key;
		if (key == KEY_BACKSPACE && i) i--;
		buffer[i] = 0;
	} 
	while (key != KEY_ENTER);
}


//--------------------------------------------------------------------------

void RefreshTile(int x, int y)
{
	char tilenum;
	
	char * sprite8_C0=NULL;
	char * sprite8_C1=NULL;
	
	int PosX = Game->Scroll_Pos.x+x;
	int PosY = Game->Scroll_Pos.y+y;
	
	if (x>=0 && x<LEVEL_WIDTH && y>=0 && y<LEVEL_HEIGHT)
	if (PosX>=0 && PosX<BOARD_WIDTH && PosY>=0 && PosY<BOARD_HEIGHT)
	{
		
		tilenum = Game->LevelMat[x][y];
		
		if (tilenum==TS_FLOOR && !ShowTileOnScreen)
		return;
		
		sprite8_C0=
		(tilenum==TS_FLOOR)         ?sprite8_0_C0:
		(tilenum==TS_WALL)          ?sprite8_1_C0:
		(tilenum==TS_GOAL)          ?sprite8_2_C0:
		(tilenum==TS_BALL)          ?sprite8_3_C0:
		(tilenum==TS_BALL_INGOAL)   ?sprite8_4_C0:
		(tilenum==TS_SOKOBAN)       ?sprite8_5_C0:
		(tilenum==TS_SOKOBAN_INGOAL)?sprite8_6_C0:NULL;
		
		sprite8_C1=
		(tilenum==TS_FLOOR)         ?sprite8_0_C1:
		(tilenum==TS_WALL)          ?sprite8_1_C1:
		(tilenum==TS_GOAL)          ?sprite8_2_C1:
		(tilenum==TS_BALL)          ?sprite8_3_C1:
		(tilenum==TS_BALL_INGOAL)   ?sprite8_4_C1:
		(tilenum==TS_SOKOBAN)       ?sprite8_5_C1:
		(tilenum==TS_SOKOBAN_INGOAL)?sprite8_6_C1:NULL;
		
		char*Screen0=ShowTileOnScreen?GetPlane(0):Plane0;
		char*Screen1=ShowTileOnScreen?GetPlane(1):Plane1;
		
		if (ShowTileOnScreen)
		GraySprite8_AND((PosX<<2)+PosX,(PosY<<2)+PosY,5,sprite8_W_C0,sprite8_W_C0,Screen0,Screen1);
		
		GraySprite8_OR((PosX<<2)+PosX,(PosY<<2)+PosY,5,sprite8_C0,sprite8_C1,Screen0,Screen1);
		
	}
}
//--------------------------------------------------------------------------

void RefreshStats(void)
{
	char buffer[50];
	
	int stats_top=0;
	
	char*Screen0=ShowTileOnScreen?GetPlane(0):Plane0;
	char*Screen1=ShowTileOnScreen?GetPlane(1):Plane1;
	
	sprintf(buffer,"File : %03d",Game->file_num);
	DrawGrayStrExt2B(STATS_LEFT,stats_top,buffer,A_REPLACE,0,Screen0,Screen1);
	stats_top+=6;
	
	sprintf(buffer,"Level : %03d",Game->level_num);
	DrawGrayStrExt2B(STATS_LEFT,stats_top,buffer,A_REPLACE,0,Screen0,Screen1);
	stats_top+=6;
	
	sprintf(buffer,"Moves : %03d",Game->moves);
	DrawGrayStrExt2B(STATS_LEFT,stats_top,buffer,A_REPLACE,0,Screen0,Screen1);
	stats_top+=6;
	
	sprintf(buffer,"Pushes : %03d",Game->pushes);
	DrawGrayStrExt2B(STATS_LEFT,stats_top,buffer,A_REPLACE,0,Screen0,Screen1);
	stats_top+=6;
	
	if (Game->saved==Game->balls)
	strcpy(buffer,"You Won!");
	else
	strcpy(buffer,"                       ");
	
	stats_top=60;
	DrawGrayStrExt2B(STATS_LEFT,stats_top,buffer,A_REPLACE,0,Screen0,Screen1);
	
}
//--------------------------------------------------------------------------


void RefreshScreen(void)
{
	int a,b;
	
	ClearGrayScreen2B(Plane0,Plane1);
	
	for (a=0;a<BOARD_WIDTH;a++)
	for (b=0;b<BOARD_HEIGHT;b++)
	RefreshTile(a-Game->Scroll_Pos.x,b-Game->Scroll_Pos.y);
	
	// VERTICAL LINE
	// 104 = 8 * 13
	char*ptr=Plane1+13;
	for (a=0;a<100;a++,ptr+=30)
	*ptr|=0x80;
	
	FastCopyScreen(Plane0,GetPlane(0));
	FastCopyScreen(Plane1,GetPlane(1));
	
}
//--------------------------------------------------------------------------

void GetLevelsOffsets(void)
{
	
	Game->levels_max=0;
	
	char*ptr=Game->VarPtr+2;
	
	for (;;)
	{
		if (*ptr=='L')
		Game->LevelsPtr[Game->levels_max++]=ptr;
		else if (*ptr=='E')
		break;
		
		ptr++;
	}
	
	
}
//--------------------------------------------------------------------------

int LoadFile (void)
{
	HSym hsym;
	
	char filename[20];
	
	sprintf(filename,"soklvl%02d",Game->file_num);
	
	
	hsym = SymFind(SYMSTR(filename));
	
	if (hsym.folder==0)
	{
		clrscr();
		printf("Error : File not found \n %s",filename);
		ngetchx();
		
		return 0;
	}
	else
	{
		
		Game->VarPtr = (char *)HeapDeref((DerefSym(hsym))->handle);
		
		clrscr();
		printf("Loading... \n");
		GetLevelsOffsets();
		printf("File : %s \n",filename);
		printf("Levels : %d \n",Game->levels_max);
		ngetchx();
		return 1;
	}
	
}
//--------------------------------------------------------------------------


void LoadLevel (void)
{
	
	
	if (Game->level_num<0)
	Game->level_num=0;
	if (Game->level_num>(Game->levels_max-1))
	Game->level_num=(Game->levels_max-1);
	
	
	Game->moves = 0;
	Game->moves_max = 0;
	
	Game->SOKO_Pos.x = 0;
	Game->SOKO_Pos.y = 0;
	
	Game->balls = 0;
	Game->pushes = 0;
	Game->saved = 0;
	
	// LOADS LEVEL
	
	memset(Game->LevelMat,TS_FLOOR,LEVEL_WIDTH*LEVEL_HEIGHT);
	
	char * ptr = Game->LevelsPtr[Game->level_num];
	ptr+=3;
	
	
	int x = 0;
	int y = 0;
	
	
	for (;;)
	{
		
		if (*ptr=='L' || *ptr=='E')
		break;
		
		if (*ptr==0x0D)
		{
			y++;
			x = 0;
			ptr+=2;
			continue;
		}
		
		Game->LevelMat[x][y] = *ptr;
		
		if (*ptr==TS_BALL)
		{
			Game->balls++;
		}
		else if (*ptr==TS_BALL_INGOAL)
		{
			Game->balls++;
			Game->saved++;
		}
		else if (*ptr==TS_SOKOBAN || *ptr==TS_SOKOBAN_INGOAL)
		{
			Game->SOKO_Pos.x = x;
			Game->SOKO_Pos.y = y;
		}
		
		x++;
		ptr++;
		
	}
	
	
}
//--------------------------------------------------------------------------


void WayToPoint (int way, TPoint * Point)
{
	
	way&=3;
	
	switch (way)
	{
		case T_LEFT :  Point->x--; break;
		case T_RIGHT : Point->x++; break;
		case T_UP :    Point->y--; break;
		case T_DOWN :  Point->y++; break;
	}
	
}
//--------------------------------------------------------------------------


void MoveSokoban(int way)
{
	
	TPoint SOKO_Dest;
	TPoint PushedBallDest;
	
	int canmove;
	
	
	SOKO_Dest=Game->SOKO_Pos;
	WayToPoint(way,&SOKO_Dest);
	
	PushedBallDest=SOKO_Dest;
	WayToPoint(way,&PushedBallDest);
	
	
	char * SOKO_MatPtr = &Game->LevelMat[SOKO_Dest.x][SOKO_Dest.y];
	
	if (SOKO_Dest.x>=0 && SOKO_Dest.x<LEVEL_WIDTH && SOKO_Dest.y>=0 && SOKO_Dest.y<LEVEL_HEIGHT)
	{
		
		canmove = 0;
		
		if ((*SOKO_MatPtr)==TS_FLOOR || (*SOKO_MatPtr)==TS_GOAL)
		{
			AddToHistory(way,0);
			canmove = 1;
		}
		
		else if ((*SOKO_MatPtr)==TS_BALL || (*SOKO_MatPtr)==TS_BALL_INGOAL)
		{
			
			char * PushedBallPtr = &Game->LevelMat[PushedBallDest.x][PushedBallDest.y];
			
			if (PushedBallDest.x>=0 && PushedBallDest.x<LEVEL_WIDTH && PushedBallDest.y>=0 && PushedBallDest.y<LEVEL_HEIGHT)
			{
				if ((*PushedBallPtr)==TS_FLOOR || (*PushedBallPtr)==TS_GOAL)
				{
					AddToHistory(way,1);
					canmove = 1;
					
					if ((*SOKO_MatPtr)==TS_BALL)
					{
						(*SOKO_MatPtr) = TS_FLOOR;
					}
					else if ((*SOKO_MatPtr)==TS_BALL_INGOAL)
					{
						(*SOKO_MatPtr) = TS_GOAL;
						Game->saved--;
					}
					
					ShowTileOnScreen=1;
					RefreshTile(SOKO_Dest.x,SOKO_Dest.y);
					ShowTileOnScreen=0;
					
					if ((*PushedBallPtr)==TS_FLOOR)
					{
						(*PushedBallPtr) = TS_BALL;
					}
					else if ((*PushedBallPtr)==TS_GOAL)
					{
						(*PushedBallPtr) = TS_BALL_INGOAL;
						Game->saved++;
					}
					
					ShowTileOnScreen=1;
					RefreshTile(PushedBallDest.x,PushedBallDest.y);
					ShowTileOnScreen=0;
					
					Game->pushes++;
				}
			}
		}
		
		if (canmove)
		{
			
			char * MatSOKOPtr = &Game->LevelMat[Game->SOKO_Pos.x][Game->SOKO_Pos.y];
			
			if ((*MatSOKOPtr)==TS_SOKOBAN)
			(*MatSOKOPtr) = TS_FLOOR;
			else if ((*MatSOKOPtr)==TS_SOKOBAN_INGOAL)
			(*MatSOKOPtr) = TS_GOAL;
			
			ShowTileOnScreen=1;
			RefreshTile(Game->SOKO_Pos.x,Game->SOKO_Pos.y);
			ShowTileOnScreen=0;
			
			if ((*SOKO_MatPtr)==TS_FLOOR)
			(*SOKO_MatPtr) = TS_SOKOBAN;
			else if ((*SOKO_MatPtr)==TS_GOAL)
			(*SOKO_MatPtr) = TS_SOKOBAN_INGOAL;
			
			ShowTileOnScreen=1;
			RefreshTile(SOKO_Dest.x,SOKO_Dest.y);
			ShowTileOnScreen=0;
			
			Game->SOKO_Pos.x = SOKO_Dest.x;
			Game->SOKO_Pos.y = SOKO_Dest.y;
			Game->moves++;
		}
	}
}
//--------------------------------------------------------------------------


void AddToHistory(int way, int moveball)
{
	char HistoryData;
	
	Game->MD_Way = (way&3);
	Game->MD_MoveBall = moveball;
	CONV_MD_HP(&HistoryData);
	Game->History[Game->moves] = HistoryData;
	Game->moves_max = Game->moves+1;
	
}
//--------------------------------------------------------------------------

void UndoMove(void)
{
	
	if (Game->moves>0)
	{
		
		char HistoryData = Game->History[Game->moves-1];
		CONV_HP_MD(&HistoryData);
		
		
		char * MatSOKOPtr = &Game->LevelMat[Game->SOKO_Pos.x][Game->SOKO_Pos.y];
		
		if ((*MatSOKOPtr)==TS_SOKOBAN)
		(*MatSOKOPtr) = TS_FLOOR;
		else
		(*MatSOKOPtr) = TS_GOAL;
		
		TPoint NewSokoPoint_P=Game->SOKO_Pos;
		WayToPoint(Game->MD_Way+0,&NewSokoPoint_P);
		TPoint NewSokoPoint_M=Game->SOKO_Pos;
		WayToPoint(Game->MD_Way+2,&NewSokoPoint_M);
		
		char * MatNewPlacePtr = &Game->LevelMat[NewSokoPoint_M.x][NewSokoPoint_M.y];
		
		if ((*MatNewPlacePtr)==TS_GOAL)
		(*MatNewPlacePtr) = TS_SOKOBAN_INGOAL;
		else
		(*MatNewPlacePtr) = TS_SOKOBAN;
		
		char * MatNewPlacePtr2 = &Game->LevelMat[NewSokoPoint_P.x][NewSokoPoint_P.y];
		
		if (Game->MD_MoveBall)
		{
			
			if ((*MatNewPlacePtr2)==TS_BALL)
			(*MatNewPlacePtr2) = TS_FLOOR;
			else
			{
				(*MatNewPlacePtr2) = TS_GOAL;
				Game->saved--;
			}
			
			if ((*MatSOKOPtr)==TS_GOAL)
			{
				(*MatSOKOPtr) = TS_BALL_INGOAL;
				Game->saved++;
			}
			else
			(*MatSOKOPtr) = TS_BALL;
			
			Game->pushes--;
		}
		
		ShowTileOnScreen=1;
		RefreshTile(NewSokoPoint_P.x,NewSokoPoint_P.y);
		RefreshTile(Game->SOKO_Pos.x,Game->SOKO_Pos.y);
		RefreshTile(NewSokoPoint_M.x,NewSokoPoint_M.y);
		ShowTileOnScreen=0;
		
		Game->moves--;
		
		Game->SOKO_Pos=NewSokoPoint_M;
		
	}
}
//--------------------------------------------------------------------------

void RedoMove(void)
{
	int moves_push;
	int moves_max_push;
	
	char HistoryData;
	
	if (Game->moves<Game->moves_max)
	{
		moves_push = Game->moves;
		moves_max_push = Game->moves_max;
		
		HistoryData = Game->History[moves_push];
		CONV_HP_MD(&HistoryData);
		
		MoveSokoban(Game->MD_Way);
		Game->moves_max = moves_max_push;
	}
	
}
//--------------------------------------------------------------------------

void CONV_HP_MD(char * HistoryPtr)
{
	Game->MD_MoveBall = (*HistoryPtr)&0x10;
	Game->MD_Way = (*HistoryPtr)&3;
}
//--------------------------------------------------------------------------

void CONV_MD_HP(char * HistoryPtr)
{
	(*HistoryPtr) = 0;
	
	if (Game->MD_MoveBall)
	(*HistoryPtr)|=0x10;
	
	(*HistoryPtr)|=(Game->MD_Way&3);
	
}
//--------------------------------------------------------------------------

void Menu_Select_Level(void)
{
	
	GrayOff();
	clrscr();
	printf("Level (0-%d) : \n",Game->levels_max-1);
	char buffer[50];
	InputStr(buffer,10);
	GrayOn();
	
	Game->level_num = atoi(buffer);
	LoadLevel();
	
	
}
//--------------------------------------------------------------------------

// Main Function
void _main(void)
{
	int key;
	
	if (!(Plane0=malloc(2*LCD_SIZE)))
	goto end_prog;
	Plane1=Plane0+LCD_SIZE;
	
	
	TMenu MenuA;
	Menu = &MenuA;
	
	TGame GameA;
	Game=&GameA;
	
	
	randomize ();
	
	
	ShowTileOnScreen=0;
	
	FontSetSys(1);
	
	
	for (;;)
	{
		Lbl_Main_Menu:
		
		MenuNew2();
		MenuAddTitle("Sokoban");
		MenuAddItem("Play");
		MenuAddItem("About");
		MenuAddItem("Quit");
		
		MenuDo(ME_MAIN);
		
		if (!Menu->ok)
		break;
		
		if (Menu->Index==1)
		{
			
			
			
			clrscr();
			printf("File Number : \n");
			char buffer[50];
			InputStr(buffer,10);
			
			Game->file_num = atoi(buffer);
			
			if (!LoadFile())
			goto Lbl_Main_Menu;
			
			
			
			Game->level_num = 0;
			LoadLevel();
			
			GrayOn();
			RefreshScreen();
			
			
			for (;;)
			{
				
				ShowTileOnScreen=1;
				RefreshStats();
				ShowTileOnScreen=0;
				
				key = ngetchx();
				
				
				
				if (key==KEY_LEFT)
				MoveSokoban(T_LEFT);
				else if (key==KEY_RIGHT)
				MoveSokoban(T_RIGHT);
				else if (key==KEY_UP)
				MoveSokoban(T_UP);
				else if (key==KEY_DOWN)
				MoveSokoban(T_DOWN);
				
				else if (key==(KEY_LEFT+(signed)KEY_SHIFT))
				{
					Game->Scroll_Pos.x++;
					RefreshScreen();
				}
				
				else if (key==(KEY_RIGHT+(signed)KEY_SHIFT))
				{
					Game->Scroll_Pos.x--;
					RefreshScreen();
				}
				
				else if (key==(KEY_UP+(signed)KEY_SHIFT))
				{
					Game->Scroll_Pos.y++;
					RefreshScreen();
				}
				
				else if (key==(KEY_DOWN+(signed)KEY_SHIFT))
				{
					Game->Scroll_Pos.y--;
					RefreshScreen();
				}
				
				else if (key==KEY_F1)
				{
					Menu_Select_Level();
					RefreshScreen();
				}
				
				else if (key==KEY_F2)
				{
					LoadLevel();
					RefreshScreen();
				}
				
				else if (key==KEY_F3)
				{
					Game->Scroll_Pos.x = 0;
					Game->Scroll_Pos.y = 0;
					RefreshScreen();
				}
				
				else if (key==KEY_F4)
				{
					Game->Scroll_Pos.x = BOARD_WIDTH/2-Game->SOKO_Pos.x;
					Game->Scroll_Pos.y = BOARD_HEIGHT/2-Game->SOKO_Pos.y;
					RefreshScreen();
				}
				
				else if (key==KEY_BACKSPACE)
				{
					UndoMove();
				}
				
				else if (key==KEY_CLEAR)
				{
					RedoMove();
				}
				
				else if (key=='+')
				{
					Game->level_num++;
					LoadLevel();
					RefreshScreen();
				}
				
				else if (key=='-')
				{
					Game->level_num--;
					LoadLevel();
					RefreshScreen();
				}
				
				else if (key=='*')
				{
					Game->level_num = random(Game->levels_max);
					LoadLevel();
					RefreshScreen();
				}
				
				else if (key==KEY_APPS)
				{
					off();
				}
				
				else if (key==KEY_ESC)
				{
					
					GrayOff();
					
					clrscr();
					printf("
[ESC] : Continue
[ENTER] : Exit");
					
					while ((key = ngetchx())!=KEY_ENTER && key!=KEY_ESC);
					
					GrayOn();
					
					if (key==KEY_ENTER)
					break;
					
					RefreshScreen();
				}
				
				
			}
			
			GrayOff();
			
		}
		
		else if (Menu->Index==2)
		{
			clrscr();
			FontSetSys(0);
			printf("About \n");
			printf("Sokoban 2.2 \n");
			printf("Programmed by Olivier Sangala \n");
			printf("http://olivier.sangala.free.fr \n");
			printf("olivier_sangala@hotmail.com \n");
			FontSetSys(1);
			ngetchx();
		}
		
		else if (Menu->Index==3)
		{
			break;
		}
		
	}
	
	end_prog:
	
	if (Plane0)
	free(Plane0);
	
}
